module Linear1LPO where

-- lex combination of linear polynomial and LPO

import Control.Monad
import Data.List

import Term
import Rule
import TRS
import DP
import Order
import SMT
import qualified Polynomial as P
import Usable
import Precedence
import Status

-- LPO

type Parameter = (P.Algebra, Precedence, Status)

showParameter :: Parameter -> String
showParameter (a, p, st) =
  P.showAlgebra a ++
  showPrecedence p ++
  showStatus st

geqLPO :: Parameter -> Term -> Term -> Bool
geqLPO _ (V x) (V y) = x == y
geqLPO _ (V _) (F _ _) = False
geqLPO param@(_, _, st) s@(F _ _) t@(V _) =
  any (\s' -> geqLPO param s' t) (args st s)
geqLPO param@(_, p, st) s@(F _ _) t@(F _ _) = 
  any (\s' -> geqLPO param s' t) (args st s) ||
  (all (\t' -> gtLPO param s t') (args st t) &&
    (gtPrec p s t || (geqPrec p s t && geqLPOLex param (args st s) (args st t))))

gtLPO :: Parameter -> Term -> Term -> Bool
gtLPO _ (V _) _ = False
gtLPO param@(_, _, st) s@(F _ _) t@(V _) =
  any (\s' -> geqLPO param s' t) (args st s)
gtLPO param@(_, p, st) s@(F _ _) t@(F _ _) = 
  any (\s' -> geqLPO param s' t) (args st s) ||
  (all (\t' -> gtLPO param s t') (args st t) &&
    (gtPrec p s t || (geqPrec p s t && gtLPOLex param (args st s) (args st t))))

-- args are assumed to be already filtered
geqLPOLex :: Parameter -> [Term] -> [Term] -> Bool
geqLPOLex _ [] [] = True
geqLPOLex _ [] (_ : _) = False
geqLPOLex _ (_ : _) []  = True
geqLPOLex param (s : ss) (t : ts) =
  gtLPO param s t || 
  (geqLPO param s t && geqLPOLex param ss ts)

gtLPOLex :: Parameter -> [Term] -> [Term] -> Bool
gtLPOLex _ [] [] = False
gtLPOLex _ [] (_ : _) = False
gtLPOLex _ (_ : _) []  = True
gtLPOLex param (s : ss) (t : ts) =
  gtLPO param s t || 
  (geqLPO param s t && gtLPOLex param ss ts)

-- combination of >_A and LPO

geqALPO :: Parameter -> Term -> Term -> Bool
geqALPO param@(a, _, _) s t = P.gtA a (s, t) || (P.geqA a (s, t) && geqLPO param s t)

gtALPO :: Parameter -> Term -> Term -> Bool
gtALPO param@(a, _, _) s t = P.gtA a (s, t) || (P.geqA a (s, t) && gtLPO param s t)

-- Encoding functions.

type Table = [(Term,Term)]

buildTable :: [Rule] -> Table
buildTable trs =
  nub [ (s', t') | (lhs, rhs) <- trs,
                   s' <- subterms lhs,
                   t' <- subterms rhs ]

varF0 :: String -> Exp
varF0 f = Var ("a_0_" ++ f) 

varFi :: String -> Int -> Formula
varFi f i = FVar ("a_" ++ show i ++ "_" ++ f) 

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

varStatus :: String -> Int -> Formula
varStatus f i = FVar ("st_" ++ show i ++ "_" ++ f) 

varGeqLPO :: Table -> Rule -> Formula
varGeqLPO table rule
  | Just i <- elemIndex rule table = FVar ("geqLPO_" ++ show i)
  | otherwise                      = error "varGeqLPO"

varGtLPO :: Table -> Rule -> Formula
varGtLPO table rule
  | Just i <- elemIndex rule table = FVar ("gtLPO_" ++ show i)
  | otherwise                      = error "varGtLPO"

constantPart :: Term -> Exp
constantPart (V _)    = Val 0
constantPart (F f ts) =
  SMT.plus (varF0 f : [ times01 (varFi f i) (constantPart ti)
                      | (i,ti) <- zip [1..] ts ])

coefficient :: String -> Term -> Exp
coefficient x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient x (F f ts) =
  SMT.plus [ times01 (varFi f i) (coefficient x ti) | (i, ti) <- zip [1..] ts ]

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA rule@(l, r) =
  conj (SMT.geq (constantPart l) (constantPart r) :
        [ SMT.geq (coefficient x l) (coefficient x r) 
        | x <- Rule.variables rule ])

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA rule@(l, r) =
  conj (SMT.gt (constantPart l) (constantPart r) :
        [ SMT.geq (coefficient x l) (coefficient x r) 
        | x <- Rule.variables rule ])

-- f >= g
encodeGeqPrec :: Rule -> Formula
encodeGeqPrec (F f _, F g _) = SMT.geq (varPrec f) (varPrec g)
encodeGeqPrec _ = error "encodeGeqPrec"

-- f > g
encodeGtPrec :: Rule -> Formula
encodeGtPrec (F f _, F g _) = SMT.gt (varPrec f) (varPrec g)
encodeGtPrec _ = error "encodeGtPrec"

-- s >=_lpo t
encodeGeqLPO :: Table -> Rule -> Formula
encodeGeqLPO _ (s@(V _), t)
  | s == t = top
  | otherwise = bottom
encodeGeqLPO table rl@(F _ _, V _) = encodeLPOSub table rl
encodeGeqLPO table rl@(F f ss, F g ts) =
  disj [
    encodeLPOSub table rl,
    conj [
      encodeLPOArgs table rl,
      disj [
        encodeGtPrec rl,
        -- only strict precedence is considered for now
        if f == g then encodeGeqLPOLex table f ss ts else bottom
      ]
    ]
  ]

-- s >_LPO t
encodeGtLPO :: Table -> Rule -> Formula
encodeGtLPO _ (V _, _) = bottom
encodeGtLPO table rl@(F _ _, V _) = encodeLPOSub table rl
encodeGtLPO table rl@(F f ss, F g ts) =
  disj [
    encodeLPOSub table rl,
    conj [
      encodeLPOArgs table rl,
      disj [
        encodeGtPrec rl,
        -- only strict precedence is considered for now
        if f == g then encodeGtLPOLex table f ss ts else bottom
      ]
    ]
  ]

encodeLPOSub :: Table -> Rule -> Formula
encodeLPOSub table (F f ss, t) =
 disj [ conj [ varStatus f i, varGeqLPO table (s', t) ] | (s', i) <- zip ss [1..] ]
encodeLPOSub _ _ = error "encodeLPOSub"

encodeLPOArgs :: Table -> Rule -> Formula
encodeLPOArgs table (s, F g ts) =
 conj [ implies (varStatus g j) (varGtLPO table (s, t')) | (t', j) <- zip ts [1..] ]
encodeLPOArgs _ _ = error "encodeLPOArgs"

encodeGtLPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGtLPOLex tbl f ss ts = encodeGtLPOLex' tbl f 1 ss ts

encodeGtLPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGtLPOLex' _ _ _ [] [] = bottom
encodeGtLPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtLPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqLPO tbl (s, t)), encodeGtLPOLex' tbl f (i+1) ss ts ] ]
encodeGtLPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqLPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGeqLPOLex tbl f ss ts = encodeGeqLPOLex' tbl f 1 ss ts

encodeGeqLPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGeqLPOLex' _ _ _ [] [] = top
encodeGeqLPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtLPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqLPO tbl (s, t)), encodeGeqLPOLex' tbl f (i+1) ss ts ] ]
encodeGeqLPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqALPO :: Table -> Rule -> Formula
encodeGeqALPO tbl rl = disj [ encodeGtA rl, conj [ encodeGeqA rl, varGeqLPO tbl rl ] ]

encodeGtALPO :: Table -> Rule -> Formula
encodeGtALPO tbl rl = disj [ encodeGtA rl, conj [ encodeGeqA rl, varGtLPO tbl rl ] ]

encode_one :: TRS -> TRS -> Formula
encode_one p r =
  conj (disj [ encodeGtALPO table rl | rl <- p ] :
        [ encodeGeqALPO table rl | rl <- p ++ r ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
         -- combinability (NOTE: accidentally equivalent to pi-simplicity?)
        [ implies (varStatus f i) (varFi f i) | (f, n) <- fs, i <- [1..n] ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtLPO table rule) (encodeGtLPO table rule) | rule <- table ] ++
        [ implies (varGeqLPO table rule) (encodeGeqLPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf p) (TRS.signatureOf r)
    table = buildTable (p ++ r)

encode_all :: TRS -> TRS -> Formula
encode_all p r =
  conj ([ encodeGtALPO table rl | rl <- p ] ++
        [ encodeGeqALPO table rl | rl <- p ++ r ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
         -- combinability (NOTE: accidentally equivalent to pi-simplicity?)
        [ implies (varStatus f i) (varFi f i) | (f, n) <- fs, i <- [1..n] ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtLPO table rule) (encodeGtLPO table rule) | rule <- table ] ++
        [ implies (varGeqLPO table rule) (encodeGeqLPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf p) (TRS.signatureOf r)
    table = buildTable (p ++ r)

-- Decoding functions

decodeInterpretation :: String -> Int -> Model -> P.Interpretation
decodeInterpretation f n model = (f, (f0, fs))
  where 
    f0 = evalExp model (varF0 f)
    fs = [ if evalFormula model (varFi f i) then 1 else 0 | i <- [1 .. n] ]

decodeAlgebra :: Signature -> Model -> P.Algebra
decodeAlgebra sig model =
  [ decodeInterpretation f n model | (f, n) <- sig ]

decodePrecedence :: Signature -> Model -> Precedence
decodePrecedence sig model =
  [ (f, evalExp model (varPrec f)) | (f, _) <- sig ]

decodeStatus :: Signature -> Model -> Status
decodeStatus sig model =
  [ (f, [ evalFormula model (varStatus f i) | i <- [1..n] ]) | (f, n) <- sig ]

decode :: TRS -> TRS -> Model -> Parameter
decode r s model =
  (decodeAlgebra sig model,
   decodePrecedence sig model,
   decodeStatus sig model)
  where
    sig =  union (TRS.signatureOf r) (TRS.signatureOf s)

solveOne :: String -> Problem -> IO (Maybe (Order, TRS, Problem))
solveOne tool (ps, rs, m)= do
  result <- sat tool (encode_one ps us)
  case result of
    Nothing -> return Nothing
    Just model 
      | length ps' < length ps ->
          return (Just (Linear1LPO a p st, us, (ps', rs, m))) 
      | otherwise -> error "solveOne"
      where 
        params@(a, p, st) = decode ps us model
        ps' = [ rule | rule@(l, r) <- ps, not (gtALPO params l r) ]
  where
    us = Usable.usable_rules ps rs

solveAll :: String -> Problem -> IO (Maybe Order)
solveAll tool (ps, rs, _)= do
  result <- sat tool (encode_all ps rs)
  case result of
    Nothing -> return Nothing
    Just model -> do
          forM_ ps checkGt
          forM_ rs checkGeq
          return (Just (Linear1LPO a p st)) 
      where 
        param@(a, p, st) = decode ps rs model
        checkGt (s, t)
          | gtALPO param s t = return ()
          | otherwise = error (show s ++ " > " ++ show t ++ " is not satisfied")
        checkGeq (s, t)
          | geqALPO param s t = return ()
          | otherwise = error (show s ++ " >= " ++ show t ++ " is not satisfied")
