module LinearWPO where

-- WPO with linear polynomial on N
-- strict precedence (not quasi-)

import Control.Monad
import Data.List

import Term
import TRS
import Rule
import SMT
import qualified Polynomial as P
import Precedence
import Status
import Order
import DP
import Usable

type Parameter = (P.Algebra, Precedence, Status)

showParameter :: Parameter -> String
showParameter (a, p, st) =
  P.showAlgebra a ++
  showPrecedence p ++
  showStatus st

geqWPO :: Parameter -> Term -> Term -> Bool
geqWPO (_, _, _) (V x) (V y) = x == y
geqWPO (a, _, _) s@(V _) t@(F _ _) = P.gtA a (s, t)
geqWPO param@(a, _, st) s@(F _ _) t@(V _) =
  P.gtA a (s, t) || (P.geqA a (s, t) && any (\s' -> geqWPO param s' t) (args st s))
geqWPO param@(a, p, st) s@(F _ _) t@(F _ _) = 
  P.gtA a (s, t) ||
    (P.geqA a (s, t) &&
      (any (\s' -> geqWPO param s' t) (args st s) ||
       (all (\t' -> gtWPO param s t') (args st t) &&
        (gtPrec p s t || (geqPrec p s t && geqWPOLex param (args st s) (args st t))))))

gtWPO :: Parameter -> Term -> Term -> Bool
gtWPO (a, _, _) s@(V _) t = P.gtA a (s, t)
gtWPO param@(a, _, st) s@(F _ _) t@(V _) =
  P.gtA a (s, t) || (P.geqA a (s, t) && any (\s' -> geqWPO param s' t) (args st s))
gtWPO param@(a, p, st) s@(F _ _) t@(F _ _) =
  P.gtA a (s, t) ||
    (P.geqA a (s, t) &&
      (any (\s' -> geqWPO param s' t) (args st s) ||
       (all (\t' -> gtWPO param s t') (args st t) &&
        (gtPrec p s t || (geqPrec p s t && gtWPOLex param (args st s) (args st t))))))

-- args are assumed to be already filtered
geqWPOLex :: Parameter -> [Term] -> [Term] -> Bool
geqWPOLex _ [] [] = True
geqWPOLex _ [] (_ : _) = False
geqWPOLex _ (_ : _) []  = True
geqWPOLex param (s : ss) (t : ts) =
  gtWPO param s t || 
  (geqWPO param s t && geqWPOLex param ss ts)

gtWPOLex :: Parameter -> [Term] -> [Term] -> Bool
gtWPOLex _ [] [] = False
gtWPOLex _ [] (_ : _) = False
gtWPOLex _ (_ : _) []  = True
gtWPOLex param (s : ss) (t : ts) =
  gtWPO param s t || 
  (geqWPO param s t && gtWPOLex param ss ts)

-- Encoding functions.

type Table = [(Term,Term)]

varF0 :: String -> Exp
varF0 f = Var ("a_0_" ++ f) 

varFi :: String -> Int -> Formula
varFi f i = FVar ("a_" ++ show i ++ "_" ++ f) 

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

varStatus :: String -> Int -> Formula
varStatus f i = FVar ("st_" ++ show i ++ "_" ++ f) 

varGeqWPO :: Table -> Rule -> Formula
varGeqWPO table rule
  | Just i <- elemIndex rule table = FVar ("geqWPO_" ++ show i)
  | otherwise                      = error "varGeqWPO"

varGtWPO :: Table -> Rule -> Formula
varGtWPO table rule
  | Just i <- elemIndex rule table = FVar ("gtWPO_" ++ show i)
  | otherwise                      = error "varGtWPO"

constantPart :: Term -> Exp
constantPart (V _)    = Val 0
constantPart (F f ts) =
  SMT.plus (varF0 f : [ times01 (varFi f i) (constantPart ti)
                      | (i,ti) <- zip [1..] ts ])

coefficient :: String -> Term -> Exp
coefficient x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient x (F f ts) =
  SMT.plus [ times01 (varFi f i) (LinearWPO.coefficient x ti) | (i, ti) <- zip [1..] ts ]

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA rule@(l, r) =
  conj (SMT.geq (constantPart l) (constantPart r) :
        [ SMT.geq (LinearWPO.coefficient x l) (LinearWPO.coefficient x r) 
        | x <- Rule.variables rule ])

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA rule@(l, r) =
  conj (SMT.gt (constantPart l) (constantPart r) :
        [ SMT.geq (LinearWPO.coefficient x l) (LinearWPO.coefficient x r) 
        | x <- Rule.variables rule ])

-- f >= g
encodeGeqPrec :: Rule -> Formula
encodeGeqPrec (F f _, F g _) = SMT.geq (varPrec f) (varPrec g)
encodeGeqPrec _ = error "encodeGeqPrec"

-- f > g
encodeGtPrec :: Rule -> Formula
encodeGtPrec (F f _, F g _) = SMT.gt (varPrec f) (varPrec g)
encodeGtPrec _ = error "encodeGtPrec"

-- s >_wpo t
encodeGtWPO :: Table -> Rule -> Formula
encodeGtWPO _ rl@(V _, _) = encodeGtA rl
encodeGtWPO table rl@(F _ _, V _) =
  disj [ encodeGtA rl, conj [ encodeGeqA rl, encodeWPOSub table rl] ]
encodeGtWPO table rl@(F f ss, F g ts) =
  disj [ encodeGtA rl,
    conj [ encodeGeqA rl,
      disj [
        encodeWPOSub table rl,
        conj [
          encodeWPOArgs table rl,
          disj [
            encodeGtPrec rl,
            -- only strict precedence is considered for now
            if f == g then encodeGtWPOLex table f ss ts else bottom
          ]
        ]
      ]
    ]
  ]


-- s >=_wpo t
encodeGeqWPO :: Table -> Rule -> Formula
encodeGeqWPO _ (V x, V y) = if x == y then top else bottom
encodeGeqWPO _ rl@(V _, _) = encodeGtA rl
encodeGeqWPO table rl@(F _ _, V _) =
  disj [ encodeGtA rl, conj [ encodeGeqA rl, encodeWPOSub table rl] ]
encodeGeqWPO table rl@(F f ss, F g ts) =
  disj [ encodeGtA rl,
    conj [ encodeGeqA rl,
      disj [
        encodeWPOSub table rl,
        conj [
          encodeWPOArgs table rl,
          disj [
            encodeGtPrec rl,
            -- only strict precedence is considered for now
            if f == g then encodeGeqWPOLex table f ss ts else bottom
          ]
        ]
      ]
    ]
  ]

encodeWPOSub :: Table -> Rule -> Formula
encodeWPOSub table (F f ss, t) =
 disj [ conj [ varStatus f i, varGeqWPO table (s', t) ] | (s', i) <- zip ss [1..] ]
encodeWPOSub _ _ = error "encodeWPOSub"

encodeWPOArgs :: Table -> Rule -> Formula
encodeWPOArgs table (s, F g ts) =
 conj [ implies (varStatus g j) (varGtWPO table (s, t')) | (t', j) <- zip ts [1..] ]
encodeWPOArgs _ _ = error "encodeWPOArgs"

encodeGtWPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGtWPOLex tbl f ss ts = encodeGtWPOLex' tbl f 1 ss ts

encodeGtWPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGtWPOLex' _ _ _ [] [] = bottom
encodeGtWPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtWPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqWPO tbl (s, t)), encodeGtWPOLex' tbl f (i+1) ss ts ] ]
encodeGtWPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqWPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGeqWPOLex tbl f ss ts = encodeGeqWPOLex' tbl f 1 ss ts

encodeGeqWPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGeqWPOLex' _ _ _ [] [] = top
encodeGeqWPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtWPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqWPO tbl (s, t)), encodeGeqWPOLex' tbl f (i+1) ss ts ] ]
encodeGeqWPOLex' _ _ _ _ _ = error "length of terms must be the same"

encode_one :: TRS -> TRS -> Formula
encode_one trs1 trs2 =
  conj (disj [ varGtWPO table rl | rl <- trs1 ] :
        [ varGeqWPO table rl | rl <- trs1 ++ trs2 ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
         -- pi-simplicity
        [ implies (varStatus f i) (varFi f i) | (f, n) <- fs, i <- [1..n] ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtWPO table rule) (encodeGtWPO table rule) | rule <- table ] ++
        [ implies (varGeqWPO table rule) (encodeGeqWPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf trs1) (TRS.signatureOf trs2)
    table = nub [ (s', t') | (lhs, rhs) <- trs1 ++ trs2,
                             s' <- subterms lhs,
                             t' <- subterms rhs ]

encode_all :: TRS -> TRS -> Formula
encode_all trs1 trs2 =
  conj ([ varGtWPO table rl | rl <- trs1 ] ++
        [ varGeqWPO table rl | rl <- trs1 ++ trs2 ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
         -- pi-simplicity
        [ implies (varStatus f i) (varFi f i) | (f, n) <- fs, i <- [1..n] ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtWPO table rule) (encodeGtWPO table rule) | rule <- table ] ++
        [ implies (varGeqWPO table rule) (encodeGeqWPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf trs1) (TRS.signatureOf trs2)
    table = nub [ (s', t') | (lhs, rhs) <- trs1 ++ trs2,
                             s' <- subterms lhs,
                             t' <- subterms rhs ]

-- Decoding functions

decodeInterpretation :: String -> Int -> Model -> P.Interpretation
decodeInterpretation f n model = (f, (f0, fs))
  where 
    f0 = evalExp model (varF0 f)
    fs = [ if evalFormula model (varFi f i) then 1 else 0 | i <- [1 .. n] ]

decodeAlgebra :: Signature -> Model -> P.Algebra
decodeAlgebra sig model =
  [ decodeInterpretation f n model | (f, n) <- sig ]

decodePrecedence :: Signature -> Model -> Precedence
decodePrecedence sig model =
  [ (f, evalExp model (varPrec f)) | (f, _) <- sig ]

decodeStatus :: Signature -> Model -> Status
decodeStatus sig model =
  [ (f, [ evalFormula model (varStatus f i) | i <- [1..n] ]) | (f, n) <- sig ]

decode :: TRS -> TRS -> Model -> Parameter
decode r s model =
  (decodeAlgebra sig model,
   decodePrecedence sig model,
   decodeStatus sig model)
  where
    sig =  union (TRS.signatureOf r) (TRS.signatureOf s)

solveOne :: String -> Problem -> IO (Maybe (Order, TRS, Problem))
solveOne tool (ps, rs, m)= do
  result <- sat tool (encode_one ps us)
  case result of
    Nothing -> return Nothing
    Just model 
      | length ps' < length ps ->
          return (Just (LinearWPO a p st, us, (ps', rs, m))) 
      | otherwise -> error "solveOne"
      where 
        params@(a, p, st) = decode ps us model
        ps' = [ rule | rule@(l, r) <- ps, not (gtWPO params l r) ]
  where
    us = Usable.usable_rules ps rs

solveAll :: String -> Problem -> IO (Maybe Order)
solveAll tool (ps, rs, _)= do
  result <- sat tool (encode_all ps rs)
  case result of
    Nothing -> return Nothing
    Just model -> do
          forM_ ps checkGt
          forM_ rs checkGeq
          return (Just (LinearWPO a p st)) 
      where 
        param@(a, p, st) = decode ps rs model
        checkGt (s, t)
          | gtWPO param s t = return ()
          | otherwise = error (show s ++ " > " ++ show t ++ " is not satisfied")
        checkGeq (s, t)
          | geqWPO param s t = return ()
          | otherwise = error (show s ++ " >= " ++ show t ++ " is not satisfied")
