module MaxPlus where

import Data.List
import qualified Polynomial as P
import qualified Term as T
import qualified Rule as R

-- (f0, [(f1, f'1), ..., (fn, f'n)]) represents
-- max{ f0, f1 + f'1 x_1, ..., fn + f'n x_n }
type MaxPlusParam = (Int, [(Int, Int)])

-- f(x1,...,xn) = max{ f0, f1 + f'1 x_1, ..., fn + f'n x_n }
type Interpretation = (String, MaxPlusParam)

type Algebra = [Interpretation]

showMonomial :: (Int, String) -> String
showMonomial (0, _) = "0"
showMonomial (1, x) = x
showMonomial (n, x) = show n ++ " " ++ x

-- pretty printer for f + f'x
showLinearPolynomial :: (Int, Int, String) -> String
showLinearPolynomial (f, 0, _x) = show f
showLinearPolynomial (0, 1, x) = x
showLinearPolynomial (f, 1, x) = show f ++ " + " ++ x
showLinearPolynomial (0, f', x) = show f' ++ x
showLinearPolynomial (f, f', x) = show f ++ " + " ++ show f' ++ " " ++ x

-- can be improved
showMaxPlus :: MaxPlusParam -> String
showMaxPlus (a0, []) = show a0
showMaxPlus (a0, as) = "max{" ++ (intercalate ", " ms) ++ "}"
  where
    m1n = [ showLinearPolynomial (ai, ai', "x" ++ show i)
          | (i, (ai, ai')) <- zip [ 1 :: Int ..] as]
    ms = show a0 : m1n

showInterpretation :: Interpretation -> String
showInterpretation (f, (a0, [])) =
  f ++ "_A = " ++ show a0
showInterpretation (f, mp@(_, as)) =
  f ++ "_A(" ++ xs ++ ") = " ++ showMaxPlus mp
  where
    xs = intercalate "," [ "x" ++ show i | i <- [1 .. length as] ]

showAlgebra :: Algebra -> String
showAlgebra interpretations =
  "max/plus interpretations on N:\n" ++
  -- show interpretations ++ -- DEBUG
  "\n" ++
  unlines 
    [ "  " ++ showInterpretation interpretation 
    | interpretation <- interpretations ] ++ "\n"

-- max{p1, ..., pn}
type MaxPlus = [P.Polynomial] -- TODO: better naming? (the same can be said for P.Polynomial)

var :: String -> MaxPlus
var x = [ P.var x ]

num :: Int -> MaxPlus
num n = [ P.num n ]

-- unfold max
flatten :: [MaxPlus] -> MaxPlus
flatten mps = concat mps

plus :: Int -> MaxPlus -> MaxPlus
plus k ps = [P.plus [(P.num k), p] | p <- ps]

times :: Int -> MaxPlus -> MaxPlus
times k ps = [ P.times k p | p <- ps]

-- max G >= max H
geq :: MaxPlus -> MaxPlus -> Bool
geq gs hs = all (\h -> (any (\g -> P.geq g h) gs)) hs

-- max G > max H
gt :: MaxPlus -> MaxPlus -> Bool
gt gs hs = all (\h -> (any (\g -> P.gt g h) gs)) hs


interpret :: Algebra -> T.Term -> MaxPlus
interpret _ (T.V x) = MaxPlus.var x 
interpret a (T.F f ts)
  | Just (a0, as) <- lookup f a = 
    MaxPlus.flatten (MaxPlus.num a0 : [ MaxPlus.plus ai (MaxPlus.times ai' (interpret a ti)) | ((ai, ai'), ti) <- zip as ts ])
interpret _ _ = error "interpret"

-- >=_A
geqA :: Algebra -> R.Rule -> Bool
geqA a (l, r) = MaxPlus.geq (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> R.Rule -> Bool
gtA a (l, r) = MaxPlus.gt (interpret a l) (interpret a r)
