module Polynomial where

import Data.List
import Data.Maybe
import qualified Rule as R
import qualified Term as T

-- f_0 + f_1 x_1 + ... + f_n x_n
type Coefficients = (Int, [Int])

-- f(x1,...,xn) = f_0 + f_1 x_1 + ... + f_n x_n
type Interpretation = (String,Coefficients)

type Algebra = [Interpretation]

showMonomial :: (Int, String) -> String
showMonomial (0, _) = "0"
showMonomial (1, x) = x
showMonomial (n, x) = show n ++ " " ++ x

-- print as polynomials
showCoefficients :: (String, Coefficients) -> String
showCoefficients (x, (a0, as)) = 
  case ms1 ++ ms2 of
    [] -> "0"
    ms -> intercalate " + " ms
  where
    ms1 = [ showMonomial (ai, x ++ show i)
          | (i, ai) <- zip [1 :: Int ..] as, ai /= 0 ]
    ms2 = if a0 == 0 then [] else [show a0]

showInterpretation :: (String, Coefficients) -> String
showInterpretation (f, (a0,[])) =
  f ++ "_A" ++ " = " ++ show a0
showInterpretation (f, p@(_,as)) =
  f ++ "_A" ++ "(" ++ xs ++ ") = " ++ showCoefficients ("x",p)
  where
    xs = intercalate "," [ "x" ++ show i | i <- [1 .. length as] ]

showAlgebra :: Algebra -> String
showAlgebra interpretationsA =
  unlines
    (["The weakly monotone algebra (N, >) with",
      ""] ++
     ["  " ++ showInterpretation (f, p) 
     | (f, p) <- interpretationsA])

-- linear polynomial
-- constant part is associated with empty string ""
-- a bit hacky, any better implementation?
type Polynomial = [(String, Int)]

var :: String -> Polynomial
-- HACK: force to have constant part (see gt)
var x = [(x,1), ("", 0)]

num :: Int -> Polynomial
num n = [("",n)]

normalize :: Polynomial -> Polynomial
normalize []           = []
normalize ((x, m) : v) =
  (x, m + sum [ n | (y, n) <- v, x == y ]) : 
  normalize [ (y, n) | (y, n) <- v, x /= y ]

plus :: [Polynomial] -> Polynomial
plus vs = normalize (concat vs)

times :: Int -> Polynomial -> Polynomial
times k v = [ (x, k * n) | (x, n) <- v ]

coefficient :: String -> Polynomial -> Int
coefficient x v = fromMaybe 0 (lookup x v)

variables :: Polynomial -> [String]
variables v = [ x | (x, _) <- v ]

geq :: Polynomial -> Polynomial -> Bool
geq v1 v2 = all p (union (variables v1) (variables v2))
  where
    p x = coefficient x v1 >= coefficient x v2
 
-- NOTE: assumption: both v1 and v2 have constant parts (associated with the empty string "")
-- otherwise X > X could hold (as polynomial)  
gt :: Polynomial -> Polynomial -> Bool
gt v1 v2 = all p (union (variables v1) (variables v2))
  where
    p "" = coefficient "" v1 >  coefficient "" v2
    p x  = coefficient x  v1 >= coefficient x  v2

interpret :: Algebra -> T.Term -> Polynomial
interpret _ (T.V x) = var x 
interpret a (T.F f ts)
  | Just (a0, as) <- lookup f a = 
    Polynomial.plus (Polynomial.num a0 : [ Polynomial.times ai (interpret a ti) | (ai, ti) <- zip as ts ])
interpret _ _ = error "interpret"

-- >=_A
geqA :: Algebra -> R.Rule -> Bool
geqA a (l, r) = Polynomial.geq (interpret a l) (interpret a r)

-- >_A
gtA :: Algebra -> R.Rule -> Bool
gtA a (l, r) = Polynomial.gt (interpret a l) (interpret a r)

-- sanity check:
-- >>> Polynomial.gtA [] (V "x", V "x")
-- False
--
